/*
 *    snake - a videogame derived from Q-Basic nibbles
 *    Copyright (C) 1999-2000 Cyberandro J.K. Starrik
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "snake.h"

#include <stdio.h>
#include <time.h>		/* time (for srand) */
#include <sys/time.h>		/* itimer ...   */
#include <signal.h>		/* signal */
#include <setjmp.h>		/* setjmp, longjmp */

/*
 * extern functions 
 */
extern int init_graphics (void);
extern void finish_graphics (void);
extern void move_snakes (void);
extern void init_fruit (void);
extern int parse_input (void);
extern void init_modules (void);
extern void draw_arena (ARENA *);
extern int parse_cmdline_opts (int, char **);
extern void parse_cmdline_args (int, char **);
extern int parse_config ();
extern int CMD_delay (char *);
extern int CMD_kill (char *);
extern int CMD_reset (char *);
extern int CMD_trim (char *);
extern int CMD_help (char *);
extern int CMD_source (char *);
extern void expire_messages (void);

#ifdef ENABLE_CMDFIFO
extern int CMD_cmdfifo (char *);
extern void init_cmdfifo (void);
extern void read_cmdfifo (void);
extern FILE *cmdfifo;
#endif

/*
 * Private functions 
 */
static void snake_play (void);
static void sigint (int);
static void sigsegv (int);
static int CMD_quit (char *);
#if ENABLE_TIMEOUT
static void sigvtalrm (int);
#endif


#define PLAYERTIMEOUT 1


/*
 * Public function 
 */
int register_player (snake * s);
void finish (void);

/*
 * Public data 
 */
snake *zoo[MAXPLAYERS];		/* snakes pointer */
int numplayers = 0;		/* number of players */
int delay = 40;			/* game delay in milliseconds */
unsigned long int turns = 0;	/* game turns */

/*
 * Private data
 */
#if ENABLE_TIMEOUT
static jmp_buf playloop;
#endif

int
main (int argc, char **argv)
{
  int optind;

#ifndef HAVE_USLEEP
  int i;
#endif

  /* register core commands */
  register_command ("help", CMD_help, "help <command>");
  register_command ("source", CMD_source, "source <filename>");
  register_command ("delay", CMD_delay, "delay <#>");
#ifdef ENABLE_CMDFIFO
  register_command ("cmdfifo", CMD_cmdfifo, "cmdfifo <path>");
#endif
  register_command ("kill", CMD_kill, "kill <snakenum>");
  register_command ("reset", CMD_reset, "reset <snakenum>");
  register_command ("trim", CMD_trim, "trim <snakenum>");
  register_command ("quit", CMD_quit, "quit");

  /* parse command line options */
  optind = parse_cmdline_opts (argc, argv);

  /*
   * Initialize graphics, arena and fruit
   */
  init_graphics ();
  arenainit ();
  init_fruit ();

  /*
   * initialize the modules before parsing command line args and config file
   * ( they may introduce new commands )
   */
  init_modules ();

  /*
   * parse configuration file 
   */
  parse_config ();

  /*
   * parse command line arguments
   * parse it after configuration file
   * to override values specified there
   */
  parse_cmdline_args (argc - optind, argv + optind);

  /*
   * Set signal handlers
   */
  signal (SIGINT, sigint);
  signal (SIGSEGV, sigsegv);
#if ENABLE_TIMEOUT
  signal (SIGVTALRM, sigvtalrm);	/* players timeout */
#endif

  /*
   * initialize the random seed 
   */
  srand (time (NULL));

  /* update the screen */
  updatescreen ();

  /* MAIN LOOP  */
  for (;;)
    {

      /* parse keyboard input */
      parse_input ();

#ifdef ENABLE_CMDFIFO
      /* Read from command fifo */
      if (cmdfifo)
	read_cmdfifo ();
#endif

      /* give players a turn to play */
      snake_play ();

      /* move the snakes */
      move_snakes ();

      /* draw the arena */
      draw_arena (&arena);

      /* update the screen with the new data */
      updatescreen ();

      /* sleep for delay milliseconds (should use setitimer) */
      //napms (delay); /* <- this requires libcurses */

#ifdef HAVE_USLEEP
      usleep(delay*1000);
#else
      for (i=0; i<30000*delay; i++);
#endif

      /* increment bogus time counter */
      turns++;

      //expire_messages ();

    }

  finish ();
  exit (0);			/* finish will exit itself */
}

void
finish ()
{
  int i;

  /*
   * reset the screen 
   */
  finish_graphics();

  printf ("\n");

  /*
   * print scores 
   */
  for (i = 0; zoo[i]; i++)
    {
      printf ("snake %i: score(%li), length(%i)\n",
	      i, zoo[i]->score, zoo[i]->length);
    }

  exit (0);
}


static void
snake_play (void)
{
  static int i; /* could be messed on longjump */

#if ENABLE_TIMEOUT

  static struct itimerval playtimer = {
    {0, PLAYERTIMEOUT},		/*
				 * it_interval.{tv_sec,tv_usec} 
				 */
    {0, PLAYERTIMEOUT},		/*
				 * it_value.{tv_sec,tv_usec} 
				 */
  };
#endif


  for (i = 0; zoo[i]; i++)
    {

#if ENABLE_TIMEOUT
      if (setjmp (playloop))
	{

	  /*
	   * reset timer 
	   */
	  setitimer (ITIMER_VIRTUAL, NULL, NULL);

	  message ("!! snake %i thinks too much !!", i);

	  continue;
	}
#endif

      if (zoo[i]->died)
	continue;

      /*
       * rewind typein queue 
       */
      rewindtypein ();

#if ENABLE_TIMEOUT
      /*
       * start timer 
       */
      setitimer (ITIMER_VIRTUAL, &playtimer, NULL);
#endif

      /*
       * call player function 
       */
      (*zoo[i]->play) (zoo[i]);

#if ENABLE_TIMEOUT
      /*
       * remove timer 
       */
      setitimer (ITIMER_VIRTUAL, NULL, NULL);
#endif

    }
}


/*
 * player modules must call this function at startup 
 */
int
register_player (snake * s)
{
  /*
   * don't register more then MAXPLAYERS snakes ! 
   * (gotta remove this limitation)
   */
  if (numplayers >= MAXPLAYERS - 1)
    {
      message ("maximum count of players reached");
      return 0;
    }

  /*
   * register new player 
   */
  zoo[numplayers] = s;

  /*
   * set numplayers flag 
   */
  zoo[numplayers]->playernum = numplayers;

  message ("player %d registered", numplayers);

  return numplayers++;
}

static void
sigint (int sig)
{
  finish ();
}

static void
sigsegv (int sig)
{
  fprintf (stderr, "SIGSEGV received, finishing...\n");
  finish ();
}

#if ENABLE_TIMEOUT
static void
sigvtalrm (int sig)
{
  /* reset signal handler (linux needs this, bsd doesn't) */
  signal (sig, sigvtalrm);
  longjmp (playloop, 1);
}
#endif


static int
CMD_quit (char *dummy)
{
  finish ();
  return 1;
}


